# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
import sys
from collections import namedtuple
from dataclasses import dataclass
from enum import IntEnum
from threading import Event
from time import sleep, monotonic

from ovos_utils.file_utils import get_temp_path
from ovos_utils.log import LOG


@dataclass
class RuntimeRequirements:
    # to ensure backwards compatibility the default values require internet before skill loading
    # skills in the wild may assume this behaviour and require network on initialization
    # any ovos aware skills should change these as appropriate

    # xxx_before_load is used by skills service
    network_before_load: bool = True
    internet_before_load: bool = True
    gui_before_load: bool = False

    # this allows a skill/plugin to spec if it needs connectivity to handle utterances
    requires_internet: bool = True
    requires_network: bool = True
    requires_gui: bool = False

    # this allows a skill/plugin to spec if it has a fallback for temporary offline events, eg, by having a cache
    no_internet_fallback: bool = False
    no_network_fallback: bool = False
    no_gui_fallback: bool = True  # can work voice only


class MonotonicEvent(Event):
    """Event class with monotonic timeout.

    Normal Event doesn't do wait timeout in a monotonic manner and may be
    affected by changes in system time. This class wraps the Event class
    wait() method with logic guards ensuring monotonic operation.
    """

    def wait_timeout(self, timeout):
        """Handle timeouts in a monotonic way.

        Repeatingly wait as long the event hasn't been set and the
        monotonic time doesn't indicate a timeout.

        Args:
            timeout: timeout of wait in seconds

        Returns:
            True if Event has been set, False if timeout expired
        """
        result = False
        end_time = monotonic() + timeout

        while not result and (monotonic() < end_time):
            # Wait however many seconds are left until the timeout has passed
            sleep(0.1)  # Mainly a precaution to not busy wait
            remaining_time = end_time - monotonic()
            LOG.debug('Will wait for {} sec for Event'.format(remaining_time))
            result = super().wait(remaining_time)

        return result

    def wait(self, timeout=None):
        if timeout is None:
            ret = super().wait()
        else:
            ret = self.wait_timeout(timeout)
        return ret


class ProcessState(IntEnum):
    """Oredered enum to make state checks easy.

    For example Alive can be determined using >= ProcessState.ALIVE,
    which will return True if the state is READY as well as ALIVE.
    """
    NOT_STARTED = 0
    STARTED = 1
    ERROR = 2
    STOPPING = 3
    ALIVE = 4
    READY = 5


# Process state change callback mappings.
_STATUS_CALLBACKS = [
    'on_started',
    'on_alive',
    'on_ready',
    'on_error',
    'on_stopping',
]

# namedtuple defaults only available on 3.7 and later python versions
if sys.version_info < (3, 7):
    StatusCallbackMap = namedtuple('CallbackMap', _STATUS_CALLBACKS)
    StatusCallbackMap.__new__.__defaults__ = (None,) * 5
else:
    StatusCallbackMap = namedtuple(
        'CallbackMap',
        _STATUS_CALLBACKS,
        defaults=(None,) * len(_STATUS_CALLBACKS),
    )


class ProcessStatus:
    """Process status tracker.

    The class tracks process status and execute callback methods on
    state changes as well as replies to messagebus queries of the
    process status.

    Args:
        name (str): process name, will be used to create the messagebus
                    messagetype "mycroft.{name}...".
        bus (MessageBusClient): Connection to the Mycroft messagebus.
        callback_map (StatusCallbackMap): optionally, status callbacks for the
                                          various status changes.
    """

    def __init__(self, name, bus=None, callback_map=None, namespace="mycroft"):
        self.name = name
        self._namespace = namespace
        self.callbacks = callback_map or StatusCallbackMap()
        self.state = ProcessState.NOT_STARTED

        # Messagebus connection
        self.bus = None
        if bus:
            self.bind(bus)

    def bind(self, bus):
        self.bus = bus
        self._register_handlers()

    def _register_handlers(self):
        """Register messagebus handlers for status queries."""
        self.bus.on(f'{self._namespace}.{self.name}.is_alive', self.check_alive)
        self.bus.on(f'{self._namespace}.{self.name}.is_ready', self.check_ready)

        # The next one is for backwards compatibility
        self.bus.on(f'mycroft.{self.name}.all_loaded', self.check_ready)

    def check_alive(self, message=None):
        """Respond to is_alive status request.

        Args:
            message: Optional message to respond to, if omitted no message
                     is sent.

        Returns:
            bool, True if process is alive.
        """
        is_alive = self.state >= ProcessState.ALIVE

        if message:
            status = {'status': is_alive}
            self.bus.emit(message.response(data=status))

        return is_alive

    def check_ready(self, message=None):
        """Respond to all_loaded status request.

        Args:
            message: Optional message to respond to, if omitted no message
                     is sent.

        Returns:
            bool, True if process is ready.
        """
        is_ready = self.state >= ProcessState.READY
        if message:
            status = {'status': is_ready}
            self.bus.emit(message.response(data=status))

        return is_ready

    def set_started(self):
        """Process is started."""
        self.state = ProcessState.STARTED
        if self.callbacks.on_started:
            self.callbacks.on_started()

    def set_alive(self):
        """Basic loading is done."""
        self.state = ProcessState.ALIVE
        if self.callbacks.on_alive:
            self.callbacks.on_alive()

    def set_ready(self):
        """All loading is done."""
        self.state = ProcessState.READY
        if self.callbacks.on_ready:
            self.callbacks.on_ready()

    def set_stopping(self):
        """Process shutdown has started."""
        self.state = ProcessState.STOPPING
        if self.callbacks.on_stopping:
            self.callbacks.on_stopping()

    def set_error(self, err=''):
        """An error has occured and the process is non-functional."""
        # Intentionally leave is_started True
        self.state = ProcessState.ERROR
        if self.callbacks.on_error:
            self.callbacks.on_error(err)


class Signal:
    """
    Capture and replace a signal handler with a user supplied function.
    The user supplied function is always called first then the previous
    handler, if it exists, will be called.  It is possible to chain several
    signal handlers together by creating multiply instances of objects of
    this class, providing a  different user functions for each instance.  All
    provided user functions will be called in LIFO order.
    """

    #
    # Constructor
    # Get the previous handler function then set the passed function
    # as the new handler function for this signal

    def __init__(self, sig_value, func):
        """
        Create an instance of the signal handler class.

        sig_value:  The ID value of the signal to be captured.
        func:  User supplied function that will act as the new signal handler.
        """
        super(Signal, self).__init__()  # python 3+ 'super().__init__()

        from signal import signal, SIG_DFL, default_int_handler, SIG_IGN

        self.__sig_value = sig_value
        self.__user_func = func  # store user passed function
        self.__previous_func = signal(sig_value, self)
        self.__previous_func = {  # Convert signal codes to functions
            SIG_DFL: default_int_handler,
            SIG_IGN: lambda a, b: None
        }.get(self.__previous_func, self.__previous_func)

    #
    # Called to handle the passed signal
    def __call__(self, signame, sf):
        """
        Allows the instance of this class to be called as a function.
        When called it runs the user supplied signal handler than
        checks to see if there is a previously defined handler.  If
        there is a previously defined handler call it.
        """
        self.__user_func()
        self.__previous_func(signame, sf)

    #
    # reset the signal handler
    def __del__(self):
        """
        Class destructor.  Called during garbage collection.
        Resets the signal handler to the previous function.
        """

        from signal import signal
        signal(self.__sig_value, self.__previous_func)


class PIDLock:  # python 3+ 'class Lock'

    """
    Create and maintains the PID lock file for this application process.
    The PID lock file is located in /tmp/mycroft/*.pid.  If another process
    of the same type is started, this class will 'attempt' to stop the
    previously running process and then change the process ID in the lock file.
    """

    @classmethod
    def init(cls):
        # TODO: Path to deprecation
        try:
            from ovos_config.meta import get_xdg_base
            base_dir = get_xdg_base()
        except ImportError:
            LOG.warning("ovos-config not available, using default "
                        "'mycroft' basedir")
            base_dir = "mycroft"
        cls.DIRECTORY = cls.DIRECTORY or get_temp_path(base_dir)

    #
    # Class constants
    DIRECTORY = None
    FILE = '/{}.pid'

    #
    # Constructor
    def __init__(self, service):
        """
        Builds the instance of this object.  Holds the lock until the
        object is garbage collected.

        service: Text string.  The name of the service application
        to be locked (ie: skills, voice)
        """
        PIDLock.init()
        super(PIDLock, self).__init__()  # python 3+ 'super().__init__()'
        self.__pid = os.getpid()  # PID of this application
        self.path = PIDLock.DIRECTORY + PIDLock.FILE.format(service)
        self.set_handlers()  # set signal handlers
        self.create()

    #
    # Reset the signal handlers to the 'delete' function
    def set_handlers(self):
        """
        Trap both SIGINT and SIGTERM to gracefully clean up PID files
        """
        from signal import SIGINT, SIGTERM
        self.__handlers = {SIGINT: Signal(SIGINT, self.delete),
                           SIGTERM: Signal(SIGTERM, self.delete)}

    #
    # Check to see if the PID already exists
    #  If it does exits perform several things:
    #    Stop the current process
    #    Delete the exiting file
    def exists(self):
        """
        Check if the PID lock file exists.  If it does
        then send a SIGKILL signal to the process defined by the value
        in the lock file.  Catch the keyboard interrupt exception to
        prevent propagation if stopped by use of Ctrl-C.
        """
        if not os.path.isfile(self.path):
            return
        with open(self.path, 'r') as L:
            try: # TODO - make it work in windows ?
                from signal import SIGKILL
                os.kill(int(L.read()), SIGKILL)
            except Exception as e:
                LOG.error(f"Failed to kill PID {L}: {e}")

    #
    # Create a lock file for this server process
    def touch(self):
        """
        If needed, create the '/tmp/mycroft' directory than open the
        lock file for writing and store the current process ID (PID)
        as text.
        """
        os.makedirs(PIDLock.DIRECTORY, exist_ok=True)
        with open(self.path, 'w') as L:
            L.write('{}'.format(self.__pid))

    #
    # Create the PID file
    def create(self):
        """
        Checks to see if a lock file for this service already exists,
        if so have it killed.  In either case write the process ID of
        the current service process to to the existing or newly created
        lock file in /tmp/mycroft/
        """
        self.exists()  # check for current running process
        self.touch()

    #
    # Delete the PID file - but only if it has not been overwritten
    # by a duplicate service application
    def delete(self, *args):
        """
        If the PID lock file contains the PID of this process delete it.

        *args: Ignored.  Required as this fuction is called as a signel
        handler.
        """
        try:
            with open(self.path, 'r') as L:
                pid = int(L.read())
                if self.__pid == pid:
                    os.unlink(self.path)
        except IOError:
            pass


def reset_sigint_handler():
    """Reset the sigint handler to the default.

    This fixes KeyboardInterrupt not getting raised when started via
    start-mycroft.sh
    """
    from signal import signal, SIGINT, default_int_handler
    signal(SIGINT, default_int_handler)



import re


class SSMLBuilder:
    def __init__(self, ssml_tag=False, speak_tag=False):
        self.text = ""
        self.ssml_tags = ["speak", "ssml", "phoneme", "voice", "audio", "sub",
                          "prosody", "break", "whisper", "p", "s", "w",
                          "emphasis"]
        self.speak_tag = speak_tag
        self.ssml_tag = ssml_tag

    def sub(self, alias=None, word=None):
        if alias is None:
            raise TypeError('Parameter alias must not be None')
        if word is None:
            raise TypeError('Parameter word must not be None')
        if len(word.strip()) == 0:
            raise ValueError('Parameter word must not be empty')
        self.text += "<sub alias='" + alias + "'>" + word + "</sub> "
        return self

    def emphasis(self, level=None, word=None):
        if level is None:
            raise TypeError('Parameter level must not be None')
        if word is None:
            raise TypeError('Parameter word must not be None')
        if len(word.strip()) == 0:
            raise ValueError('Parameter word must not be empty')
        level = level.lower().strip()
        self.text += "<emphasis level='" + level + "'>" + word + "</emphasis> "
        return self

    def parts_of_speech(self, word=None, role=None):
        """Special considerations when speaking word include usage or role of word"""
        if word is None:
            raise TypeError('Parameter word must not be None')
        if role is None:
            raise TypeError('Parameter role must not be None')
        self.text += "<w role='" + role + "'>" + word + "</w> "
        return self

    def pause_by_strength(self, strength=None):
        if strength is None:
            raise TypeError('Parameter strength must not be None')
        try:
            strength = strength.lower().strip()
        except AttributeError:
            raise AttributeError('Parameter strength must be a string')

        self.text += "<break strength=" + strength + "/>"
        return self

    def sentence(self, text=None):
        """Wrap text with <s> tag"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<s>" + text + "</s> "
        return self

    def say_emphasis(self, text=None):
        self.emphasis("strong", text)
        return self

    def say_strong(self, text=None):
        """Wrap text with <s> tag"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<amazon:effect vocal-tract-length=\"+20%\">" + text \
                     + "</amazon:effect>"
        return self

    def say_weak(self, text=None):
        """Wrap text with <s> tag"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<amazon:effect vocal-tract-length=\"-20%\">" + text \
                     + "</amazon:effect> "
        return self

    def say_softly(self, text=None):
        """Wrap text with <s> tag"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<amazon:effect phonation=\"soft\">" + text + "</amazon:effect> "
        return self

    def say_auto_breaths(self, text=None):
        """Wrap text with <s> tag"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<amazon:auto-breaths>" + text + "</amazon:auto-breaths>"
        return self

    def paragraph(self, text=None):
        """Wrap text with <p> tag"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<p>" + text + "</p> "
        return self

    def audio(self, audio_file=None, text=None):
        if audio_file is None:
            raise TypeError('Parameter audio_file must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += '<audio src=' + audio_file + '>' + text + '</audio>'
        return self

    def pause(self, time=0, unit="ms"):
        if unit not in ["s", "ms"]:
            raise TypeError("time must be in seconds or miliseconds")
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        if time > 0:
            self.text += "<break time=" + str(time) + unit + "/>"
        else:
            self.text += "<break />"
        return self

    def prosody(self, attribute=None, text=None):
        if attribute is None:
            raise TypeError('Parameter attribute must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<prosody " + attribute + ">" + text + "</prosody> "
        return self

    def pitch(self, pitch=None, text=None):
        if pitch is None:
            raise TypeError('Parameter pitch must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<prosody pitch='" + str(
            pitch) + "'>" + text + "</prosody> "
        return self

    def volume(self, volume=None, text=None):
        if volume is None:
            raise TypeError('Parameter volume must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<prosody volume='" + volume + "'>" + text + "</prosody> "
        return self

    def rate(self, rate=None, text=None):
        if rate is None:
            raise TypeError('Parameter rate must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<prosody rate='" + str(rate) + "'>" + text + \
                     "</prosody>"
        return self

    def say(self, text=None):
        """Add normal speed text to SSML"""
        if text is None:
            raise TypeError('Parameter text must not be None')
        self.rate("1", text)
        return self

    def say_loud(self, text=None):
        if text is None:
            raise TypeError('Parameter text must not be None')
        self.volume("1.6", text)
        return self

    def say_slow(self, text=None):
        if text is None:
            raise TypeError('Parameter text must not be None')
        self.rate("0.4", text)
        return self

    def say_fast(self, text=None):
        if text is None:
            raise TypeError('Parameter text must not be None')
        self.rate("1.6", text)
        return self

    def say_low_pitch(self, text=None):
        if text is None:
            raise TypeError('Parameter text must not be None')
        self.pitch("-10%", text)
        return self

    def say_high_pitch(self, text=None):
        if text is None:
            raise TypeError('Parameter text must not be None')
        self.pitch("+10%", text)
        return self

    def say_whispered(self, text=None):
        self.whisper(text)
        return self

    def phoneme(self, ph=None, text=None):
        if ph is None:
            raise TypeError('Parameter ph must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<phoneme ph=" + ph + ">" + text + "</phoneme> "
        return self

    def voice(self, voice=None, text=None):
        if voice is None:
            raise TypeError('Parameter voice must not be None')
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<voice name=" + voice + ">" + text + "</voice> "
        return self

    def whisper(self, text=None):
        if text is None:
            raise TypeError('Parameter text must not be None')
        if len(self.text) and not self.text.endswith(" "):
            self.text += " "
        self.text += "<whispered>" + text + "</whispered> "
        return self

    def build(self):
        self.text = self.text.strip()
        if self.speak_tag:
            self.text = "<speak>\n" + self.text + "\n</speak>"
        if self.ssml_tag:
            self.text = "<ssml>\n" + self.text + "\n</ssml>"
        return self.text

    @staticmethod
    def remove_ssml(text):
        return re.sub('<[^>]*>', '', text).replace('  ', ' ')

    @staticmethod
    def extract_ssml_tags(text):
        # find ssml tags in string
        return re.findall('<[^>]*>', text)



import json
from copy import copy


def nested_get(base, key_list):
    """Access a nested object in base by item sequence."""
    if not len(key_list):
        return None
    val = copy(base)
    for key in key_list:
        if key not in val:
            return None
        val = val[key]
    return val


def nested_set(base, key_list, value):
    """Set a value in a nested object in base by item sequence."""
    for key in key_list[:-1]:
        base = base.setdefault(key, {})
    base[key_list[-1]] = value
    return base


def nested_delete(base, key_list):
    """Delete a value in a nested object in base by item sequence."""
    if not len(key_list):
        return base
    d = base
    for key in key_list[:-1]:
        if key not in base:
            return base
        d = d[key]
    if key_list[-1] in d:
        d.pop(key_list[-1])
    return base


def flatten_dict(base, separator=":"):
    new_dict = {}
    for key, value in base.items():
        if isinstance(value, dict):
            new_dict.update(
                {separator.join([key, k]): v
                 for k, v in flatten_dict(value).items()}
            )
        else:
            new_dict[key] = value
    return new_dict


def flattened_get(base, key, separator=":"):
    return flatten_dict(base, separator=separator).get(key)


def flattened_set(base, key, value, separator=":"):
    keys = key.split(separator)
    return nested_set(base, keys, value)


def flattened_delete(base, key, separator=":"):
    keys = key.split(separator)
    return nested_delete(base, keys)


def invert_dict(base):
    return {v: k for k, v in base.items()}


def merge_dict(base, delta, merge_lists=False, skip_empty=False,
               no_dupes=True, new_only=False):
    """
        Recursively merges two dictionaries
        including dictionaries within dictionaries.

        Args:
            base:  Target for merge
            delta: Dictionary to merge into base
            merge_lists: if a list is found merge contents instead of replacing
            skip_empty: if an item in delta is empty, dont overwrite base
            no_dupes: when merging lists deduplicate entries
            new_only: only merge keys not yet in base
    """

    for k, d in delta.items():
        b = base.get(k)
        if isinstance(d, dict) and isinstance(b, dict):
            merge_dict(b, d, merge_lists, skip_empty, no_dupes, new_only)
        else:
            if new_only and k in base:
                continue
            if skip_empty and d in [None, "", []]:
                # dont replace if new entry is empty
                # False and 0 should still be replaced
                pass
            elif all((isinstance(b, list), isinstance(d, list), merge_lists)):
                if no_dupes:
                    base[k] += [item for item in d if item not in base[k]]
                else:
                    base[k] += d
            else:
                base[k] = d
    return base


def load_commented_json(filename):
    """ Loads an JSON file, ignoring comments

    Supports a trivial extension to the JSON file format.  Allow comments
    to be embedded within the JSON, requiring that a comment be on an
    independent line starting with '//' or '#'.

    NOTE: A file created with these style comments will break strict JSON
          parsers.  This is similar to but lighter-weight than "human json"
          proposed at https://hjson.org

    Args:
        filename (str):  path to the commented JSON file

    Returns:
        obj: decoded Python object
    """
    with open(filename) as f:
        contents = f.read()

    return json.loads(uncomment_json(contents))


def uncomment_json(commented_json_str):
    """ Removes comments from a JSON string.

    Supporting a trivial extension to the JSON format.  Allow comments
    to be embedded within the JSON, requiring that a comment be on an
    independent line starting with '//' or '#'.

    Example...
       {
         // comment
         'name' : 'value'
       }

    Args:
        commented_json_str (str):  a JSON string

    Returns:
        str: uncommented, legal JSON
    """
    lines = commented_json_str.splitlines()
    # remove all comment lines, starting with // or #
    nocomment = []
    for line in lines:
        stripped = line.lstrip()
        if stripped.startswith("//") or stripped.startswith("#"):
            continue
        nocomment.append(line)

    return " ".join(nocomment)


def is_compatible_dict(base, delta):
    """
    returns False if any key common to base/delta has a different type,
    except for None values, dicts are evaluated recursively
    """
    common_keys = [k for k in base if k in delta]
    for k in common_keys:
        if base[k] is None or delta[k] is None:
            continue
        elif isinstance(base[k], dict) and isinstance(delta[k], dict):
            if not is_compatible_dict(delta[k], base[k]):
                return False
        elif type(base[k]) != type(delta[k]):
            return False
    return True



import os
from os.path import join, isdir
from typing import List

from ovos_utils.log import LOG
from ovos_bus_client.util import wait_for_reply
from ovos_utils.system import is_installed, has_screen, is_process_running

_default_gui_apps = (
    "ovos-gui-app",
    "ovos-shell",
    "mycroft-gui-app",
    "mycroft-embedded-shell"
)


def can_display() -> bool:
    """
    Return true if a display is available
    """
    return bool(has_screen())


def is_gui_installed(applications: List[str] = _default_gui_apps) -> bool:
    """
    Return true if a GUI application is installed
    @param applications: list of applications to check for
    """
    return any((is_installed(app) for app in applications))


def is_gui_running(applications: List[str] = _default_gui_apps) -> bool:
    """
    Return true if a GUI application is running
    @param applications: list of applications to check for
    """
    return any((is_process_running(app) for app in applications))


def is_gui_connected(bus=None) -> bool:
    """
    Check if a GUI is connected to the MessageBus.
    sends "gui.status.request" and waits for "gui.status.request.response"
    @param bus: MessageBusClient to use for query
    @return: True if GUI is connected
    """
    response = wait_for_reply("gui.status.request",
                              "gui.status.request.response", bus=bus)
    if response:
        return response.data["connected"]
    return False


def can_use_local_gui() -> bool:
    """
    Returns True if a local GUI is installed and running (does not check if the
    GUI is connected to an accessible GUI service).
    """
    if can_display() and is_gui_installed() and is_gui_running():
        return True
    return False


def can_use_gui(bus=None,
                local: bool = False) -> bool:
    """
    Check if a GUI is available to connect to
    @param bus: MessageBusClient to use for query
    @param local: If True, only check for a GUI on the local host
    @return: True if a GUI is available
    """
    if local:
        return can_use_local_gui()
    return can_use_local_gui() or is_gui_connected(bus)


def get_ui_directories(root_dir: str) -> dict:
    """
    Get a dict of available UI directories by GUI framework.
    @param root_dir: base directory to inspect for available UI directories
    @return: Dict of framework name to UI resource directory
    """
    ui_directories = dict()
    if isdir(f"{root_dir}/ui"):
        LOG.debug("legacy UI directory found - Handling `ui` directory as `qt5`")
        ui_directories["qt5"] = f"{root_dir}/ui"
    elif isdir(f"{root_dir}/gui"):
        for framework in os.listdir(f"{root_dir}/gui"):
            if isdir(f"{root_dir}/gui/{framework}"):
                LOG.debug(f"Skill supports GUI framework: {framework} from folder: {root_dir}/gui/{framework}")
                ui_directories[framework] = f"{root_dir}/gui/{framework}"
    return ui_directories



from ovos_bus_client.util import wait_for_reply


def get_non_properties(obj):
    """Get attributes that are not properties from object.

    Will return members of object class along with bases down to MycroftSkill.

    Args:
        obj: object to scan

    Returns:
        Set of attributes that are not a property.
    """

    def check_class(cls):
        """Find all non-properties in a class."""
        # Current class
        d = cls.__dict__
        np = [k for k in d if not isinstance(d[k], property)]
        # Recurse through base classes excluding MycroftSkill and object
        for b in [b for b in cls.__bases__ if b.__name__ not in ("object", "MycroftSkill")]:
            np += check_class(b)
        return np

    return set(check_class(obj.__class__))


def skills_loaded(bus=None) -> bool:
    """
    Await a reply from mycroft.skills.all_loaded to check if all skills are
    loaded.
    @param bus: OVOS messagebus client
    @return: Are all skills loaded? True/False
    """
    if bus is None:
        return False
    reply = wait_for_reply('mycroft.skills.all_loaded',
                           'mycroft.skills.all_loaded.response',
                           bus=bus)
    if reply:
        return reply.data['status']
    return False



import collections
import csv
import os
import re
import tempfile
from os import walk
from os.path import dirname, splitext, join
from sys import platform
from threading import RLock
from typing import Optional, List

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from ovos_utils.bracket_expansion import expand_template
from ovos_utils.log import LOG, log_deprecation


def ensure_directory_exists(directory, domain=None):
    """ Create a directory and give access rights to all

    Args:
        domain (str): The IPC domain.  Basically a subdirectory to prevent
            overlapping signal filenames.

    Returns:
        str: a path to the directory
    """
    if domain:
        directory = os.path.join(directory, domain)

    # Expand and normalize the path
    directory = os.path.normpath(directory)
    directory = os.path.expanduser(directory)

    if not os.path.isdir(directory):
        try:
            save = os.umask(0)
            os.makedirs(directory, 0o777)  # give everyone rights to r/w here
        except OSError:
            LOG.warning("Failed to create: " + directory)
            pass
        finally:
            os.umask(save)

    return directory


def to_alnum(skill_id: str) -> str:
    """
    Convert a skill id to only alphanumeric characters
     Non-alphanumeric characters are converted to "_"

    Args:
        skill_id (str): identifier to be converted
    Returns:
        (str) String of letters
    """
    return ''.join(c if c.isalnum() else '_' for c in str(skill_id))


def get_temp_path(*args) -> str:
    """
    Generate a valid path in the system temp directory.

    This method accepts one or more strings as arguments. The arguments are
    joined and returned as a complete path inside the systems temp directory.
    Importantly, this will not create any directories or files.

    Example usage: get_temp_path('mycroft', 'audio', 'example.wav')
    Will return the equivalent of: '/tmp/mycroft/audio/example.wav'

    Args:
        path_element (str): directories and/or filename

    Returns:
        (str) a valid path in the systems temp directory
    """
    try:
        path = os.path.join(tempfile.gettempdir(), *args)
    except TypeError:
        raise TypeError("Could not create a temp path, get_temp_path() only "
                        "accepts Strings")
    return path


def get_cache_directory(folder: str) -> str:
    """
    Get a temporary cache directory, preferably in RAM.
    Note that only Linux use RAM.
    @param folder: base path to use for cache
    @return: valid cache path
    """
    path = get_temp_path(folder)
    if platform == 'linux':
        try:
            from memory_tempfile import MemoryTempfile
            path = join(MemoryTempfile(fallback=True).gettempdir(), folder)
        except ImportError:
            pass
        except Exception as e:
            LOG.exception(e)
    os.makedirs(path, exist_ok=True)
    return path


def resolve_ovos_resource_file(res_name: str,
                               extra_res_dirs: list = None) -> Optional[str]:
    """
    Convert a resource into an absolute filename.
    used internally for ovos resources
    """
    # First look for fully qualified file (e.g. a user setting)
    if os.path.isfile(res_name):
        return res_name

    if extra_res_dirs:
        for res_dir in extra_res_dirs:
            filename = join(res_dir, res_name)
            if os.path.isfile(filename):
                return filename

    # now look in bundled ovos-utils resources
    filename = join(dirname(__file__), "res", res_name)
    if os.path.isfile(filename):
        return filename

    # let's look in ovos_workshop if it's installed
    # (default skill resources live here)
    try:
        import ovos_workshop
        core_root = dirname(ovos_workshop.__file__)
        filename = join(core_root, "res", res_name)
        if os.path.isfile(filename):
            return filename
    except ImportError:
        pass

    # let's look in ovos_gui if it's installed
    # (default GUI resources live here)
    try:
        import ovos_gui
        core_root = dirname(ovos_gui.__file__)
        filename = join(core_root, "res", res_name)
        if os.path.isfile(filename):
            return filename
    except ImportError:
        pass

    # let's look in mycroft/ovos-core if it's installed
    # (default core resources live here / backwards compat)
    # TODO - remove me soon, spams deprecation logs
    try:
        import mycroft
        core_root = dirname(mycroft.__file__)
        filename = join(core_root, "res", res_name)
        if os.path.isfile(filename):
            return filename
    except ImportError:
        pass

    return None  # Resource cannot be resolved


def resolve_resource_file(res_name: str, root_path: Optional[str] = None,
                          config: dict = None) -> Optional[str]:
    """
    Convert a resource into an absolute filename.

    Resource names are in the form: 'filename.ext'
    or 'path/filename.ext'

    The system wil look for ~/.mycroft/res_name first, and
    if not found will look at /opt/mycroft/res_name,
    then finally it will look for res_name in the 'mycroft/res'
    folder of the source code package.

    Example:
    With mycroft running as the user 'bob', if you called
        resolve_resource_file('snd/beep.wav')
    it would return either '/home/bob/.mycroft/snd/beep.wav' or
    '/opt/mycroft/snd/beep.wav' or '.../mycroft/res/snd/beep.wav',
    where the '...' is replaced by the path where the package has
    been installed.

    Args:
        res_name (str): a resource path/name
        root_path: Optional root path to check
        config (dict): mycroft.conf, to read data directory from
    Returns:
        str: path to resource or None if no resource found
    """
    if config is None:
        log_deprecation(f"Expected a dict config and got None.", "0.1.0")
        try:
            from ovos_config.config import read_mycroft_config
            config = read_mycroft_config()
        except ImportError:
            LOG.warning("Config not provided and ovos_config not available")
            config = dict()

    # First look for fully qualified file (e.g. a user setting)
    if os.path.isfile(res_name):
        return res_name

    # Now look for ~/.mycroft/res_name (in user folder)
    filename = os.path.expanduser("~/.mycroft/" + res_name)
    if os.path.isfile(filename):
        return filename

    # Next look for /opt/mycroft/res/res_name
    data_dir = os.path.expanduser(config.get('data_dir', "/opt/mycroft"))
    filename = os.path.expanduser(os.path.join(data_dir, res_name))
    if os.path.isfile(filename):
        return filename

    # look in ovos_utils package itself
    found = resolve_ovos_resource_file(res_name)
    if found:
        return found

    return None  # Resource cannot be resolved


def read_vocab_file(path: str) -> List[List[str]]:
    """
    Read voc file.

    This reads a .voc file, stripping out empty lines comments and expand
    parentheses. It returns each line as a list of all expanded
    alternatives.

    Args:
        path (str): path to vocab file.

    Returns:
        List of Lists of strings.
    """
    vocab = []
    with open(path, 'r', encoding='utf8') as voc_file:
        for line in voc_file.readlines():
            if line.startswith('#') or line.strip() == '':
                continue
            vocab.append(expand_template(line.lower()))
    return vocab


def load_regex_from_file(path: str, skill_id: str) -> List[str]:
    """
    Load regex from file
    The regex is sent to the intent handler using the message bus

    Args:
        path:       path to vocabulary file (*.voc)
        skill_id:   skill_id to the regex is tied to
    """
    from ovos_utils.intents.intent_service_interface import munge_regex

    regexes = []
    if path.endswith('.rx'):
        with open(path, 'r', encoding='utf8') as reg_file:
            for line in reg_file.readlines():
                if line.startswith("#"):
                    continue
                LOG.debug('regex pre-munge: ' + line.strip())
                regex = munge_regex(line.strip(), skill_id)
                LOG.debug('regex post-munge: ' + regex)
                # Raise error if regex can't be compiled
                try:
                    re.compile(regex)
                    regexes.append(regex)
                except Exception as e:
                    LOG.warning(f'Failed to compile regex {regex}: {e}')

    return regexes


def load_vocabulary(basedir: str, skill_id: str) -> dict:
    """
    Load vocabulary from all files in the specified directory.

    Args:
        basedir (str): path of directory to load from (will recurse)
        skill_id: skill the data belongs to
    Returns:
        dict with intent_type as keys and list of list of lists as value.
    """

    vocabs = {}
    for path, _, files in walk(basedir):
        for f in files:
            if f.endswith(".voc"):
                vocab_type = to_alnum(skill_id) + splitext(f)[0]
                vocs = read_vocab_file(join(path, f))
                if vocs:
                    vocabs[vocab_type] = vocs
    return vocabs


def load_regex(basedir: str, skill_id: str) -> List[List[str]]:
    """
    Load regex from all files in the specified directory.

    Args:
        basedir (str): path of directory to load from
        skill_id (str): skill identifier
    """
    regexes = []
    for path, _, files in walk(basedir):
        for f in files:
            if f.endswith(".rx"):
                regexes += load_regex_from_file(join(path, f), skill_id)
    return regexes


def read_value_file(filename: str, delim: str) -> collections.OrderedDict:
    """
    Read value file.

    The value file is a simple csv structure with a key and value.

    Args:
        filename (str): file to read
        delim (str): csv delimiter

    Returns:
        OrderedDict with results.
    """
    result = collections.OrderedDict()

    if filename:
        with open(filename) as f:
            reader = csv.reader(f, delimiter=delim)
            for row in reader:
                # skip blank or comment lines
                if not row or row[0].startswith("#"):
                    continue
                if len(row) != 2:
                    continue

                result[row[0]] = row[1]
    return result


def read_translated_file(filename: str, data: dict) -> Optional[List[str]]:
    """
    Read a file inserting data.

    Args:
        filename (str): file to read
        data (dict): dictionary with data to insert into file

    Returns:
        list of lines.
    """
    if filename:
        with open(filename) as f:
            text = f.read().replace('{{', '{').replace('}}', '}')
            return text.format(**data or {}).rstrip('\n').split('\n')
    else:
        return None


class FileWatcher:
    def __init__(self, files: List[str], callback: callable,
                 recursive: bool = False, ignore_creation: bool = False):
        """
        Initialize a FileWatcher to monitor the specified files for changes
        @param files: list of paths to monitor for file changes
        @param callback: function to call on file change with modified file path
        @param recursive: If true, recursively include directory contents
        @param ignore_creation: If true, ignore file creation events
        """
        self.observer = Observer()
        self.handlers = []
        for file_path in files:
            if os.path.isfile(file_path):
                watch_dir = dirname(file_path)
            else:
                watch_dir = file_path
            self.observer.schedule(FileEventHandler(file_path, callback,
                                                    ignore_creation),
                                   watch_dir, recursive=recursive)
        self.observer.start()

    def shutdown(self):
        """
        Remove observer scheduled events and stop the observer.
        """
        self.observer.unschedule_all()
        self.observer.stop()


class FileEventHandler(FileSystemEventHandler):
    def __init__(self, file_path: str, callback: callable,
                 ignore_creation: bool = False):
        """
        Create a handler for file change events
        @param file_path: file_path being watched Unused(?)
        @param callback: function to call on file change with modified file path
        @param ignore_creation: if True, only track file modification events
        """
        super().__init__()
        self._callback = callback
        self._file_path = file_path
        if ignore_creation:
            self._events = ('modified')
        else:
            self._events = ('created', 'modified')
        self._changed_files = []
        self._lock = RLock()

    def on_any_event(self, event):
        if event.is_directory:
            return
        with self._lock:
            if event.event_type == "closed":
                if event.src_path in self._changed_files:
                    self._changed_files.remove(event.src_path)
                    # fire event, it is now safe
                    try:
                        self._callback(event.src_path)
                    except:
                        LOG.exception("An error occurred handling file "
                                      "change event callback")

            elif event.event_type in self._events:
                if event.src_path not in self._changed_files:
                    self._changed_files.append(event.src_path)



from datetime import datetime
from typing import Any

from dateutil.tz import gettz, tzlocal

# used to calculate timespans
DAYS_IN_1_YEAR = 365.2425
DAYS_IN_1_MONTH = 30.42


def get_config_tz() -> Any:
    """Get the configured timezone or, if missing, defaults to local timezone

    Returns:
        Any: timezone
    """
    try:
        from ovos_config.locale import get_config_tz as _get_config_tz
        return _get_config_tz()
    except ImportError:
        return tzlocal()


def now_utc() -> datetime:
    """ Retrieve the current time in UTC

    Returns:
        (datetime): The current time in Universal Time, aka GMT
    """
    return datetime.utcnow().replace(tzinfo=gettz("UTC"))


def now_local(tz: datetime.tzinfo = None) -> datetime:
    """ Retrieve the current time

    Args:
        tz (datetime.tzinfo, optional): Timezone, default to user's settings

    Returns:
        (datetime): The current time
    """
    tz = tz or get_config_tz()
    return datetime.now(tz)


def to_utc(dt: datetime) -> datetime:
    """ Convert a datetime with timezone info to a UTC datetime

    Args:
        dt (datetime): A datetime (presumably in some local zone)
    Returns:
        (datetime): time converted to UTC
    """
    tz = gettz("UTC")
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=get_config_tz())
    return dt.astimezone(tz)


def to_local(dt: datetime) -> datetime:
    """ Convert a datetime to the user's local timezone

    Args:
        dt (datetime): A datetime (if no timezone, defaults to UTC)
    Returns:
        (datetime): time converted to the local timezone
    """
    tz = get_config_tz()
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=get_config_tz())
    return dt.astimezone(tz)


def to_system(dt: datetime) -> datetime:
    """Convert a datetime to the system's local timezone

    Args:
        dt (datetime): A datetime (if no timezone, assumed to be UTC)
    Returns:
        (datetime): time converted to the operation system's timezone
    """
    tz = tzlocal()
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=get_config_tz())
    return dt.astimezone(tz)


def is_leap_year(year: int) -> bool:
    """Checks if input year is a leap year

    Args:
        year (int): the year to check
    Returns:
        bool: if the input is an leap year
    """
    return (year % 400 == 0) or ((year % 4 == 0) and (year % 100 != 0))


def get_next_leap_year(year: int) -> int:
    """Get the following leap year of a reference year 

    Args:
        year (int): reference year

    Returns:
        int: following leap year
    """
    next_year = year + 1
    if is_leap_year(next_year):
        return next_year
    else:
        return get_next_leap_year(next_year)
